/*
** Really mindless text compression program.
**
** The intended use of the compression is to be able to fit
** E-text books (Gutenberg, OBI, Wiretap) into a smaller space
** so that they can be read on a machine with very limited store,
** such as a handheld computer.  One implication of this is that
** the form of compression does not need to be completely lossless
** as long as the result is still readable.
**
** This program assumes a linear address space at least large enough
** to load the whole text.  One implication of this is that it won't
** be usable as a simple DOS program, but the alternative was to spend
** more time on this than I felt I could afford.
*/
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>

#define NEARLY 10              /* less than this is "nearly" unused */

#define GRAM 2                 /* size of strings to scan for */
#define GRAMS 4000             /* max size of dictionary of scanned strings */

#define FALSE 0
#define TRUE 1

#define NL '\n'
#define CR '\r'

typedef unsigned char SYMBOL;

/***********************************************************************
************************************************************************
***                                                                  ***
***                   E R R O R   H A N D L I N G                    ***
***                                                                  ***
************************************************************************
***********************************************************************/

/*
** abandon
**
** Single point of exit from the program in the event of problems.
*/
static void abandon ( const char *s )
{
   fprintf(stderr, "zvrz abandoned: %s\n", s);
   exit(EXIT_FAILURE);
}

/***********************************************************************
************************************************************************
***                                                                  ***
***                    O P T I O N   H A N D L I N G                 ***
***                                                                  ***
************************************************************************
***********************************************************************/

/*
** Definition of the getopt() function and its friends.
*/
#include "getopt.h"

/*
** line_limit
**
** Sets the maximum size of an uncompressed line in the file.
*/
static int line_limit = 250;

/*
** checking
**
** If set, perform extra consistency checking.
*/
static int checking = FALSE;

/*
** resubstitute
**
** If set, resubstitute symbols that are not as good as the
** next compression.
*/
static int resubstitute = FALSE;

/*
** filtering
**
** If set, preprocess text.  If not set, don't preprocess text;
** this should give lossless compression.
*/
static int filtering = TRUE;

/*
** quiet
**
** If set, inhibits most diagnostic information.
*/
static int quiet = FALSE;

/*
** usage
**
** A useful function which indicates the correct way
** of calling this program.
*/
static void usage(void)
{
   printf("ZVRZ V1.2 by Ian A Young <ian@rats.demon.co.uk>\n\n");
   printf("zvrz -d file...         decompress to standard output\n");
   printf("zvrz [options] file...  compress files .txt -> .zvr\n\n");
   printf("   -c                   more internal checking\n");
   printf("   -l n                 set output line length\n");
   printf("   -n                   don't filter the input\n");
   printf("   -q                   operate quietly\n");
   printf("   -r                   allow resubstitutions\n");
   printf("   -t                   issue timing statistics\n");
   exit(EXIT_SUCCESS);
}

/***********************************************************************
************************************************************************
***                                                                  ***
***                        U T I L I T I E S                         ***
***                                                                  ***
************************************************************************
***********************************************************************/

/*
** pchar
**
** Prints a character in such a way that it is always readable,
** if necessary hexifying it.
*/
static void pchar (int n)
{
   if (isprint(n)) {
      putchar(n);
   } else {
      printf("<%d>", n);
   }
}

/*
** pstring
**
** Prints a given string "safely" through pchar.
*/
static void pstring ( const unsigned char *s )
{
   while (*s) pchar(*s++);
}

/***********************************************************************
************************************************************************
***                                                                  ***
***                      T E X T   S T O R A G E                     ***
***                                                                  ***
************************************************************************
***********************************************************************/

/*
** "original_size" is the size, in characters, of the original file.
** "text" is a pointer to a heap area of a little bigger than this size,
** so that it can be terminated by a couple of special characters and
** a zero if required.
*/
static size_t original_size;
static unsigned char *text;

/*
** "uncompressed_size" is the number of characters in the uncompressed,
** but filtered, text.
*/
static size_t uncompressed_size;

/*
** "current_size" is the number of characters in "text" which are
** currently valid.
*/
static size_t current_size;

/*
** process_char
**
** Called once for each character in the input file.  This is a state
** machine which weeds out things like redundant white space and
** unwanted hyphenation.
*/
static unsigned char *pp;   /* pointer through which process_char writes */

static enum state {
   END_PARA,          /* have just seen end of a paragraph         */
   END_LINE,          /* have just seen end of a line              */
   IN_WORD,           /* we are inside a word                      */
   IN_LINE,           /* inside a line, but not inside a word      */
   PEND_SPACE,        /* in white space                            */
   PEND_HYPHEN,       /* just seen a hyphen, may be at end of word */
   PEND_HYPHEN_NL     /* floop-<NL> seen                           */
} state;

static void process_char ( SYMBOL c )
{
   switch (state) {

   line_ended:
   case END_LINE:
      if (c == NL) {
         /*
         ** If we see a second newline after the end of an
         ** ordinary line, we have come to the end of the
         ** paragraph.  Issue the second newline and move
         ** into the END_PARA state.
         */
         *pp++ = c;
         state = END_PARA;
      } else if (c == ' ') {
         /*
         ** If we see a leading space after a newline, we have
         ** see the _other_ kind of paragraph-starting convention,
         ** and we must change it to the first kind.
         */
         *pp++ = NL;
         state = END_PARA;
      } else if (isalpha(c)) {
         *pp++ = c;
         state = IN_WORD;
      } else {
         *pp++ = c;
         state = IN_LINE;
      }
      break;

   case END_PARA:
      if (c == ' ') {
         /*
         ** Ignore spaces at this point; we know we are at the end of
         ** a paragraph already.
         */
      } else if (c == NL) {
         /*
         ** Ignore extra blank lines
         */
      } else {
         /*
         ** Something interesting: consider in line context
         */
         goto in_line;
      }
      break;

   in_line:
   case IN_LINE:
      state = IN_LINE;
      if (c == ' ') {
         state = PEND_SPACE;
      } else if (c == NL) {
         *pp++ = NL;
         state = END_LINE;
      } else if (isalpha(c)) {
         *pp++ = c;
         state = IN_WORD;
      } else {
         *pp++ = c;
      }
      break;

   case IN_WORD:
      if (isalpha(c)) {
         *pp++ = c;
      } else if (c == '-') {
         state = PEND_HYPHEN;
      } else {
         goto in_line;
      }
      break;

   case PEND_HYPHEN:
      if (isalpha(c)) { /* hyphen within a word */
         *pp++ = '-';
         *pp++ = c;
         state = IN_WORD;
      } else if (c == ' ') {
         /* absorb spaces after in-word hyphens */
      } else if (c == NL) {
         state = PEND_HYPHEN_NL;
      } else {
         *pp++ = '-'; /* flush hyphen */
         goto in_line;
      }
      break;

   case PEND_HYPHEN_NL:
      if (isalpha(c)) {
         /*
         ** We have seen a <alpha><-><wsp><NL><wsp><alpha>
         ** sequence in a word.  Drop the middle of it and
         ** join the two lines together, which will also
         ** join the two parts of the word together.
         */
         *pp++ = c;
         state = IN_WORD;
      } else if (c == ' ') {
         /*
         ** Absorb spaces at beginning of lines like these
         */
      } else {
         *pp++ = '-';
         *pp++ = NL;
         state = END_LINE;
         goto line_ended;
      }
      break;

   case PEND_SPACE:
      if (c == ' ') {
         /* ignore multiple white space */
      } else if (c == NL) {
         /*
         ** White space before line end: take as line end.
         */
         *pp++ = NL;
         state = END_LINE;
      } else {
         *pp++ = ' ';
         goto in_line;
      }
      break;

   default:
      fprintf(stderr, "state: %d character: %d\n", state, c);
      abandon("process state machine failure");
   }
}

static void load_file(FILE *fin)
{
   /*
   ** Work out the size of the file by reading it all in, character
   ** at a time.  Slow but sure.
   */
   original_size = 0;
   for (;;) {
      int c = fgetc(fin);
      if (c == EOF) break;
      original_size++;
   }

   if (!quiet)
      printf("original size: %ld characters\n", (long)original_size);

   /*
   ** Allocate some space for the file...
   */
   text = malloc(original_size+10);
   if (!text) abandon("could not allocate buffer for text");

   /*
   ** Reset the file pointer to the start and load the file in.
   */
   fseek(fin, 0, SEEK_SET);
   pp = text;
   state = END_PARA;
   for (;;) {
      int c = fgetc(fin);
      if (c == CR) continue; /* it's a DOS format file */
      if (c == EOF) break;
      if (filtering) {
         process_char((SYMBOL)c);
      } else {
         *pp++ = c;
      }
   }

   /*
   ** Purge state machine if we have one.
   */
   if (filtering) {
      process_char(NL);
      process_char(NL);
      process_char(NL);
   }
   current_size = pp - text;
}

/*
** replace_string
**
** Replace an arbitrary string with a shorter string replacement
** throughout the current version of the text.
*/
static int replace_string ( const unsigned char *s,
                            const unsigned char *r )
{
   int n = 0;
   unsigned char *beg = text;
   int rlen = strlen(r);
   int cut = strlen(s)-strlen(r);  /* number of characters to lose */
   text[current_size] = 0; /* terminate current text */
   while (beg = strstr(beg, s)) {
      memcpy(beg, r, rlen); /* perform replacement */
      n++;
      if (cut) memmove(beg+rlen, beg+rlen+cut, strlen(beg+rlen+cut)+1);
      current_size -= cut;
   }
   return n;
}

static unsigned char *xstrchr ( const unsigned char *x, unsigned char c )
{
   while (*x) {
      if (*x == c) return (unsigned char *)x;
      x++;
   }
   return NULL;
}

/*
** resubstitute_character
**
** Replaces a character throughout the current text with its
** (usually longer) definition.  We assume that this will not make the
** text bigger than it _originally_ was, so it is safe to move
** things aside to do this.
*/
static void resubstitute_character ( SYMBOL c, const SYMBOL *r )
{
   int rlen = strlen(r);
   int grow = strlen(r)-1;
   unsigned char *beg = text;
   text[current_size] = 0; /* terminate current text */
   while (beg = xstrchr(beg, c)) {
      if (grow) memmove(beg+grow, beg, strlen(beg)+1);
      memcpy(beg, r, rlen);
      current_size += grow;
      beg += rlen; /* don't resubstitute recursively */
   }
}

/***********************************************************************
************************************************************************
***                                                                  ***
***              C H A R A C T E R   C O U N T I N G                 ***
***                                                                  ***
************************************************************************
***********************************************************************/

static unsigned long character_count[256];

static void show_nearlies(void)
{
   int nearlies;
   int c;
   for (c=1, nearlies=0; c<256; c++) {
      int count = character_count[c];
      if (count && (count<NEARLY)) nearlies++;
   }

   if (!quiet) {
      printf("nearly empty character slots: %d\n", nearlies);
      if (nearlies) {
         printf("   ");
         for (c=1; c<256; c++) {
            if (character_count[c] == 0) continue;
            if (character_count[c] < NEARLY) pchar(c);
         }
         putchar(NL);
      }
   }
}

static int count_characters(void)
{
   int c, empties;
   unsigned char *p = text;

   /*
   ** First, zero the count array.
   */
   for (c=1; c<256; c++)
      character_count[c] = 0;

   /*
   ** Make sure we don't by accident use anything as a symbol
   ** that the Psion IOREAD function will regard as significant.
   **
   ** (1000 is large enough to prevent these turning up as "nearlies").
   */
   character_count[26] = 1000;
   character_count[NL] = 1000;
   character_count[CR] = 1000;

   while (p != (text+current_size)) {
      character_count[*p]++;
      p++;
   }

   for (c=1, empties=0; c<256; c++) {
      int count = character_count[c];
      if (count == 0) empties++;
   }

   return empties;
}

/***********************************************************************
************************************************************************
***                                                                  ***
***               S Y M B O L   D E F I N I T I O N S                ***
***                                                                  ***
************************************************************************
***********************************************************************/

/*
** definition
**
** This is a table of definitions for the 256 possible character values.
** Note that character 0 is never used for anything, in order to avoid
** C string termination problems.
**
** For other characters, the table entry is a pointer to a heap string
** which is what that character would expand to if it appeared in the
** compressed string.
**
** Note that this is NOT a recursive concept; the definition is a flattened
** string of characters.  This, if the character "q" only appears before
** "u" and is chosen as symbol "x", then the definition of "x" will be
** "qu" and "q" itself may then be redefined as something else later on.
*/
static unsigned char *definition[256];

/*
** basic_def
**
** This is a handy table of definitions of the compression symbols
** as themselves.
*/
static unsigned char *basic_def[256];

/*
** initialise_definitions
**
** Set up the basic definition array with pointers to heap strings
** defining the symbol as equivalent to itself.
*/
static void initialise_definitions(void)
{
   int c;
   basic_def[0] = "!!Compressed!!";
   for (c=1; c<256; c++) {
      unsigned char x[2];
      x[0] = c; x[1] = 0;
      basic_def[c] = strdup(x);
      if (!basic_def[c]) abandon("can't initialise basic definition");
   }
}

/*
** reset_definition
**
** Resets a particular definition to mean the character itself.
*/
void reset_definition ( SYMBOL c )
{
   if (definition[c] && definition[c] != basic_def[c]) free(definition[c]);
   definition[c] = basic_def[c];
}

static void dump_table ( FILE *fout )
{
   int ch;
   for (ch=0; ch<256; ch++) {
      const unsigned char *def = definition[ch];
      if (strlen(def)==1 && *def == ch) {
         /*
         ** If the character is defined as itself, just put out
         ** a blank line.
         */
         fputc(NL, fout);
      } else {
         /*
         ** Otherwise, put out the definition of the character
         ** on a line.
         */
         fprintf(fout, "%s\n", definition[ch]);
      }
   }
}

/*
** all_chars_literal
**
** TRUE if all characters in the given string are currently defined
** as themselves.
*/
static int all_chars_literal ( const char *p )
{
   while (*p) {
      unsigned char s = *p++;
      if (definition[s] != basic_def[s]) return FALSE;
   }
   return TRUE;
}

/*
** definition_value
**
** Returns the amount by which the given symbol definition
** makes the file smaller.
*/
static long definition_value ( SYMBOL sym )
{
   const unsigned char *d = definition[sym];

   /*
   ** If the symbol is defined as itself, we must assume it is
   ** VERY VERY valuable, so as not to try and play with it.
   */
   if (d == basic_def[sym]) return LONG_MAX;

   /*
   ** Similarly, a symbol is very valuable if any of the
   ** characters in its definition are not literals.
   */
   if (!all_chars_literal(d)) return LONG_MAX;

   /*
   ** Otherwise, it is the amount the file shrinks using it, less
   ** the cost of having the definition.
   */
   return (character_count[sym] * (strlen(d)-1)) - strlen(d);
}

/*
** return_resubstitutions
**
** Returns all substitutions whose value is less than
** some threshold.
*/
static int return_resubstitutions ( long cost )
{
   int many = 0;
   int c;

   /*
   ** If we are not really keen on resubstitutions, only
   ** do them if we actually make the document smaller by
   ** doing so.
   */
   if (!resubstitute) cost = 0; /* has to have negative cost */

   for (c=1; c<256; c++) {
      long rc = definition_value((SYMBOL)c);
      if (rc < cost) {
         unsigned char *d = definition[c];            
         if (!quiet)
            printf("   resubstituted %d cost %ld was \"%s\"\n", c, rc, d);
         resubstitute_character((SYMBOL)c, d);
         reset_definition((SYMBOL)c);
         many++;
      }
   }

   return many;
}

/*
** decompressed_size
**
** Returns the size that the text will become once decompressed.
*/
static size_t decompressed_size ( void )
{
   size_t r = 0;
   unsigned char *p = text;
   text[current_size] = 0; /* terminate current text */
   while (*p) {
      r += strlen(definition[*p++]);
   }
   return r;
}

/***********************************************************************
************************************************************************
***                                                                  ***
***                    G R A M   H A N D L I N G                     ***
***                                                                  ***
************************************************************************
***********************************************************************/

/*
** struct gram
**
** Note that sort_grams assumes that ch is the first field.
*/
struct gram {
   unsigned char ch[GRAM];
   unsigned long nrefs;
};

static struct gram grams[GRAMS];
static int cur_grams;
static unsigned long dropped_grams;

static void pgram ( const struct gram *g )
{
   int i;
   for (i=0; i<GRAM; i++) pchar(g->ch[i]);
}

static int sort_nrefs ( const void *L, const void *R )
{
   const struct gram *LL = L;
   const struct gram *RR = R;
   if (LL->nrefs < RR->nrefs) return -1;
   if (LL->nrefs > RR->nrefs) return 1;
   return 0;
}

static int sort_grams ( const void *L, const void *R )
{
   #if GRAM == 2
      /*
      ** Special case for GRAM == 2 assumes that the characters
      ** of the gram occupy the first bytes of the structure.
      ** They can therefore be treated as an unsigned short; this
      ** should even work on machines with alignment constraints
      ** as the structure itself must be appropriately aligned in
      ** that case.
      */
      unsigned short ll = *(unsigned short *)L;
      unsigned short rr = *(unsigned short *)R;
      if (ll < rr) return -1;
      if (ll > rr) return 1;
      return 0;
   #else
      /*
      ** General case: string comparison.
      */
      const struct gram *LL = L;
      const struct gram *RR = R;
      return strncmp(LL->ch, RR->ch, GRAM);
   #endif
}

/*
** sort_grams_by_gram
**
** Sort the gram table into string comparison order.
*/
static void sort_grams_by_gram(void)
{
   qsort(grams, cur_grams, sizeof(struct gram), sort_grams);
}

/*
** sort_grams_by_refs
**
** Sort the gram table into ascending order of number of references.
*/
static void sort_grams_by_refs(void)
{
   qsort(grams, cur_grams, sizeof(struct gram), sort_nrefs);
}

/*
** most_popular_gram
**
** Returns a pointer to the gram in the gram table which has the
** most references.
*/
static struct gram *most_popular_gram(void)
{
   unsigned long best = 0;
   struct gram *best_gram = grams;
   int i;
   for (i=0; i<cur_grams; i++) {
      if (grams[i].nrefs >= best) {
         best = grams[i].nrefs;
         best_gram = &grams[i];
      }
   }
   return best_gram;
}

static void count_gram ( const char *g )
{
   struct gram gram;
   struct gram *q;
   #if GRAM == 2
      gram.ch[0] = g[0];
      gram.ch[1] = g[1];
   #else
      memcpy(gram.ch, g, GRAM);
   #endif
   q = bsearch(&gram, grams, cur_grams, sizeof(struct gram), sort_grams);
   if (q) {
      q->nrefs++;
   } else if (cur_grams < GRAMS) {
      /*
      ** A gram we haven't seen before, but which we do have room
      ** for in the table.
      */
      gram.nrefs = 1;

      /*
      ** Add this gram in at the end of the table, then re-sort to
      ** put it in its proper place.
      */
      grams[cur_grams++] = gram;
      sort_grams_by_gram();
   } else {
      dropped_grams++;
   }
}

/*
** valid_gram
**
** True if the string is an allowed n-gram.  False if it
** contains NL and is therefore prohibited.
*/
#if GRAM == 2
   #define valid_gram(p) ((*p!=NL)&&(p[1]!=NL))
#else
   static int valid_gram ( const unsigned char *p )
   {
      int i;
      for (i=0; i<GRAM; i++)
         if (*p++ == NL) return FALSE;
      return TRUE;
   }
#endif

/*
** count_grams
**
** Run through the current text counting each n-gram
** as we go.
*/
static void count_grams(void)
{
   /* 123456789012345           */
   /* Quick Brown Fox           */
   /* ^             ^ (2-grams) */
   /* ^            ^  (3-grams) */
   /* p          plim           */

   unsigned char *p = text;
   unsigned char *plim = text+current_size-GRAM+1;
   dropped_grams = 0;
   while (p < plim) {
      if (valid_gram(p))
         count_gram(p);
      p++;
   }
}

static void reset_grams(void)
{
   int i;
   if (cur_grams == GRAMS) {
      int cut = cur_grams/100;
      if (cut) {   
         sort_grams_by_refs();
         #if defined(DEBUG)
            printf("cutting from %d to %d\n", cur_grams, cur_grams-cut);
         #endif
         memmove(grams, grams+cut, (cur_grams-cut)*sizeof(struct gram));
         cur_grams -= cut;
         sort_grams_by_gram();
      }
   }
   for (i=0; i<cur_grams; i++)
      grams[i].nrefs = 0;
}

static void replace_gram(int gramno, struct gram *g)
{
   unsigned char *from = text;
   unsigned char *to = text;
   unsigned char *org_limit = text + current_size;
   unsigned char *limit = text+current_size-GRAM+1;
   current_size = 0;
   while (from < limit) {
      if (!strncmp(from, g->ch, GRAM)) {
         *to++ = gramno;
         from += GRAM;
      } else {
         *to++ = *from++;
      }
      current_size++;
   }
   while (from != org_limit) {
      *to++ = *from++;
      current_size++;
   }
}

static void expand_gram ( unsigned char *buf, struct gram *g )
{
   int i;
   *buf = 0;
   for (i=0; i<GRAM; i++)
      strcat(buf, definition[g->ch[i]]);
}

static int compressions = 0;
static int decompressions = 0;
static int recombinations = 0;

/*
** compression_step()
**
** Works out what the next best compression step is, and does it.
**
** Returns TRUE if it did something useful, FALSE if we're all done.
*/
static int compression_step(void)
{
   long benefit;
   unsigned char ch;
   struct gram *g;
   unsigned char x[100];

   count_characters();

   if (!quiet) {
      printf("size now %ld (down %2.1f%%) ",
         (long)current_size,
         100.0*(original_size-current_size)/(double)original_size);
      fflush(stdout);
   }

   reset_grams();
   count_grams();

   /*
   ** Work out the definition we would want to use.
   */
   g = most_popular_gram();
   expand_gram(x, g);
   benefit = g->nrefs - strlen(x);

   if (!quiet) {
      printf("top %d-gram is \"", GRAM);
      pgram(g);
      printf("\" (benefit %ld", benefit);
      if (dropped_grams) printf(", drop %lu", dropped_grams);
      printf(")\n", g->nrefs, dropped_grams);
   }

   /*
   ** Return any previous compression symbols which are not as
   ** valuable as the one we are about to compress.  This will happen
   ** if the symbol has later been partially combined with something
   ** else.
   */
   if (return_resubstitutions(benefit)) {
      /*
      ** Having returned some symbols, it is possible that some of
      ** the text returned could be absorbed by some other symbols.
      */
      int i;
      int repl = 0;

      /*
      ** Go from older symbols (in general) to newer ones...
      ** they are more likely to be effective.
      */
      for (i=255; i; i--) {
         unsigned char *d = definition[i];
         if (d == basic_def[i]) continue;
         if (all_chars_literal(d)) {
            SYMBOL x[2];
            x[0] = i;
            x[1] = 0;
            repl += replace_string(d, x);
         }
      }
      if (!quiet && repl) printf("   with %d recombinations\n", repl);
      decompressions++;
      recombinations += repl;
      return TRUE;
   }

   /*
   ** If the best we can do is a gram which doesn't
   ** benefit us at all, just exit now.
   */
   if (benefit <= 0) return FALSE; /* run out of things to compress */

   /*
   ** Otherwise, find an unused compression symbol and
   ** make use of it for this n-gram.
   **
   ** DON'T do this if the symbol in question is going to be
   ** too long for the decompressor.
   */
   if (strlen(x) <= 15) {
      for (ch=255; ch>=1; ch--) {
         if (!character_count[ch]) {
            /*
            ** perform the replacement 
            */
            replace_gram(ch, g);
            g->nrefs = 0;

            reset_definition(ch);
            definition[ch] = strdup(x);

            if (!quiet) {
               printf("   character value %d now expands to \"", ch);
               pstring(definition[ch]);
               putchar('"');
               putchar(NL);
            }

            compressions++;
            return TRUE; /* managed to do something */
         }
      }
   }

   /*
   ** Here would be a good place to try a last-ditch attempt to
   ** extend any symbol if it is always followed by some other
   ** symbol, even if it only appears once, in the hope that the
   ** other symbol might then get freed up.
   */

   /*
   ** Run out of things to try...
   */
   return FALSE; /* we ran out of compression symbols */
}

/*
** optimise_lines
**
** This function scans the text and re-formats all the paragraphs
** to give maximal line sizes.  The effect of this is to allow the
** compression algorithms to work slightly more effectively because
** word endings are more often seen as with a space (which can be
** involved in compression) than before (NL can not be part of
** a compression symbol).
**
** Note that any such change leaves the text exactly the same
** size as it was before.
*/
static void optimise_lines(void)
{
   unsigned char *p;

   /*
   ** Terminate the current text.
   */
   text[current_size] = 0;

   /*
   ** Phase 1
   **
   ** Make each paragraph (where paragraphs are defined as being
   ** separated by blank lines) be a single line with spaces
   ** separating words.
   */
   p = text;
   while (p = strchr(p, NL)) {
      if (p[1] == NL) {
         /*
         ** THIS newline is the end of the paragraph, the NEXT
         ** one is the required blank line.  We don't want to
         ** touch this, and THEN we want to start at the start
         ** of the next paragraph.
         */
         p += 2;
      } else if (p[1] == 0) {
         /*
         ** Last newline in the file.
         */
         break;
      } else {
         /*
         ** Simple end of line, but not end of paragraph.
         */
         *p++ = ' ';
      }
   }

   /*
   ** Phase 2
   **
   ** For each paragraph/line in the file, reduce it to lines shorter
   ** than or equal to line_limit characters by converting spaces to NLs.
   */
   p = text;
   while (*p) {
      unsigned char *end; /* points to end of this line (NL character) */

      /*
      ** "p" points to the start of the current line.
      ** If this is a blank line, just skip over it.
      */
      if (*p == NL) {
         p++;
         continue;
      }

      /*
      ** We are at the start of a non-empty line.
      ** Locate it's NL character and thus figure out
      ** how long the line is.
      */
      end = strchr(p, NL);

      /*
      ** Reduce this line, while it is still too large.
      */
      while ( (end-p) > line_limit ) {
         unsigned char *brk = p+line_limit; /* candidate breakpoint */
         while (*brk != ' ') brk--;         /* run back to previous space */
         *brk = NL;                         /* overwrite with newline */
         p = brk+1;                     /* new beginning of line */
      }

      p = end+1;                        /* move to start of next line */
   }
}

void process_file ( const char *infname )
{
   char outfname[100];
   FILE *fin, *fout;
   int n;

   /*
   ** Construct the name for the output file.
   */
   strcpy(outfname, infname);
   memcpy(outfname+strlen(outfname)-3, "zvr", 3);

   printf("processing %s to %s...\n", infname, outfname);   
   fin = fopen(infname, "r");
   if (!fin) abandon("could not open input file");
   
   load_file(fin);
   fclose(fin);

   /*
   ** Perform some text replacements.
   **
   ** For "Terminal Compromise":
   **   ^[ ^H ^\   ->  e-grave
   **   ^[ ^B ^\   ->  e-acute
   **   <130>      ->  e-acute
   **
   ** General textual:
   **   Spaced out ellipsis ". . ." to "..."
   **   Compact lines of stars, lines of hyphens
   */
   if (filtering) {
      n = replace_string("\x1B\x08\x1C", "\x8A");
      if (!quiet && n) printf("e-grave replacements: %d\n", n);

      n = replace_string("\x1B\x02\x1C", "\x82");
      n += replace_string("<130>", "\x82");
      if (!quiet && n) printf("e-acute replacements: %d\n", n);

      n = replace_string(". .", "..");
      n += replace_string("...", "..");   /* two is enough for anyone */
      if (!quiet && n) printf("ellipsis compressions: %d\n", n);

      n = replace_string("* *", "**");
      n += replace_string("***", "**");   /* two is enough for anyone */
      if (!quiet && n) printf("star compressions: %d\n", n);

      /*
      ** Hyphens.  Reduce lines of hyphens to a pair of hyphens.
      ** Remove leading and trailing space from hyphen pairs.
      ** Convert double hyphen to a Psion long dash (194, 0xC4).
      */
      n = replace_string("---", "--");
      #if 0 /* not so sure about these ones */
         n += replace_string(" --", "--");
         n += replace_string("-- ", "--");
      #endif
      n += replace_string("--", "\xc4");
      if (!quiet && n) printf("hyphen compressions: %d\n", n);
   }

   /*
   ** Reformat paragraphs
   */
   if (filtering) optimise_lines();

   /*
   ** We now know how big the uncompressed text will be.
   */
   uncompressed_size = current_size;

   for (n=0; n<256; n++) reset_definition((SYMBOL)n);

   if (!quiet) printf("empty character slots: %d\n", count_characters());

   cur_grams = 0;
   while (compression_step()) {
      /* keep going while something has been done */
      if (checking) {
         if (uncompressed_size != decompressed_size()) {
            printf("un %ld de %ld\n", (long)uncompressed_size,
                                    (long)decompressed_size());
            abandon("un/de mismatch");
         }
      }
   }

   if (!quiet) {
      printf("empty character slots: %d\n", count_characters());
      show_nearlies();

      printf("compressions:     %d\n", compressions);
      printf("decompressions:   %d\n", decompressions);
      printf("recombinations:   %d\n", recombinations);
   }

   /*
   ** Create the output file.
   */
   fout = fopen(outfname, "wb");
   if (!fout) abandon("could not open output file");
   dump_table(fout);
   fwrite(text, 1, current_size, fout);
   fclose(fout);

   /*
   ** Clean up behind us.
   */
   free(text);
}

void decompress_file ( const char *fname )
{
   FILE *fin = fopen(fname, "rt");
   unsigned char line[1000];
   unsigned char *def[256];
   int i;

   if (!fin) abandon("could not open input file");

   fgets(line, sizeof(line), fin);
   if (strcmp(line, "!!Compressed!!\n")) {
      printf("not a compressed file\n");
      exit(EXIT_FAILURE);
   }

   /*
   ** Read rest of dictionary
   */
   for (i=1; i<256; i++) {
      fgets(line, sizeof(line), fin);
      if (line[0] == '\n') {
         line[0] = i;
         line[1] = 0;
      } else {
         line[strlen(line)-1] = 0; /* trim off newline */
      }
      def[i] = strdup(line);
   }

   /*
   ** Process the file
   */
   while (fgets(line, sizeof(line), fin)) {
      unsigned char *s = line;
      while (*s) printf("%s", def[*s++]);
   }

   /*
   ** Tidy up
   */
   fclose(fin);
   for (i=1; i<256; i++) free(def[i]);
}

/*
** main
*/
int main(int argc, const char *argv[])
{
   int timing = FALSE;
   int decompressing = FALSE;
   int errflg = FALSE;
   int c;

   /*
   ** If no arguments, give help.
   */
   if (argc == 1) usage();

   initialise_definitions();

   /*
   ** Parse command line options
   */
   while (!errflg && (c = getopt(argc, argv, "cdl:nqrt")) != -1) {
      switch (c) {

      case 'c':
         checking = TRUE;
         break;

      case 'd':
         decompressing = TRUE;
         break;

      case 'l':
         line_limit = atoi(optarg);
         if (line_limit < 40) {
            printf("line limit too small: reset\n");
            line_limit = 40;
         } else if (line_limit > 250) {
            printf("line limit too large: reset\n");
            line_limit = 250;
         }
         break;

      case 'n':
         filtering = FALSE;
         break;

      case 'q':
         quiet = TRUE;
         break;

      case 'r':
         resubstitute = TRUE;
         break;

      case 't':
         timing = TRUE;
         break;

      default:
         errflg = TRUE;
      }
   }

   if (errflg) abandon("error in option parsing");

   while (optind != argc) {
      if (decompressing) {
         decompress_file(argv[optind]);
      } else {
         time_t t1, t2;
         t1 = time(NULL);
         process_file(argv[optind]);
         t2 = time(NULL);
         if (timing) printf("   elapsed time: %.2f hours\n",
            difftime(t2, t1)/3600.0);
      }
      optind++;
   }
   
   return EXIT_SUCCESS;
}
